// C program to delete a node from AVL Tree

#include "AVL.h"

// A utility function to right rotate subtree rooted with y
// See the diagram given above.
NodePointer rightRotate(NodePointer y)
{
	NodePointer x = y->leftChild;
	NodePointer T2 = x->rightChild;

	// Perform rotation
	x->rightChild = y;
	y->leftChild = T2;

	// Update heights
	y->height = getMaxValue(getTreeHeight(y->leftChild), getTreeHeight(y->rightChild))+1;
	x->height = getMaxValue(getTreeHeight(x->leftChild), getTreeHeight(x->rightChild))+1;

	// Return new root
	return x;
}

// A utility function to left rotate subtree rooted with x
// See the diagram given above.
NodePointer leftRotate(NodePointer x)
{
	NodePointer y = x->rightChild;
	NodePointer T2 = y->leftChild;

	// Perform rotation
	y->leftChild = x;
	x->rightChild = T2;

	// Update heights
	x->height = getMaxValue(getTreeHeight(x->leftChild), getTreeHeight(x->rightChild))+1;
	y->height = getMaxValue(getTreeHeight(y->leftChild), getTreeHeight(y->rightChild))+1;

	// Return new root
	return y;
}

// Get Balance factor of node N
int getBalance(NodePointer N)
{
	if (N == NULL)
		return 0;
	return getTreeHeight(N->leftChild) - getTreeHeight(N->rightChild);
}

NodePointer insert(NodePointer node, int key)
{
	/* 1. Perform the normal BST rotation */
	if (node == NULL)
		return(createNewNode(key));

	if (key < node->data.key)
		node->leftChild = insert(node->leftChild, key);
	else if (key > node->data.key)
		node->rightChild = insert(node->rightChild, key);
	else // Equal keys not allowed
		return node;

	/* 2. Update height of this ancestor node */
	node->height = 1 + getMaxValue(getTreeHeight(node->leftChild),
						getTreeHeight(node->rightChild));

	/* 3. Get the balance factor of this ancestor
		node to check whether this node became
		unbalanced */
	int balance = getBalance(node);

	// If this node becomes unbalanced, then there are 4 cases

	// Left Left Case
	if (balance > 1 && key < node->leftChild->data.key)
		return rightRotate(node);

	// Right Right Case
	if (balance < -1 && key > node->rightChild->data.key)
		return leftRotate(node);

	// Left Right Case
	if (balance > 1 && key > node->leftChild->data.key)
	{
		node->leftChild = leftRotate(node->leftChild);
		return rightRotate(node);
	}

	// Right Left Case
	if (balance < -1 && key < node->rightChild->data.key)
	{
		node->rightChild = rightRotate(node->rightChild);
		return leftRotate(node);
	}

	/* return the (unchanged) node pointer */
	return node;
}

/* Given a non-empty binary search tree, return the
node with minimum key value found in that tree.
Note that the entire tree does not need to be
searched. */
NodePointer minValueNode(NodePointer node)
{
	NodePointer current = node;

	/* loop down to find the leftmost leaf */
	while (current->leftChild != NULL)
		current = current->leftChild;

	return current;
}

// Recursive function to delete a node with given key
// from subtree with given root. It returns root of
// the modified subtree.
NodePointer deleteNode(NodePointer root, int key)
{
	// STEP 1: PERFORM STANDARD BST DELETE

	if (root == NULL)
		return root;

	// If the key to be deleted is smaller than the
	// root's key, then it lies in left subtree
	if ( key < root->data.key )
		root->leftChild = deleteNode(root->leftChild, key);

	// If the key to be deleted is greater than the
	// root's key, then it lies in right subtree
	else if( key > root->data.key )
		root->rightChild = deleteNode(root->rightChild, key);

	// if key is same as root's key, then This is
	// the node to be deleted
	else
	{
		// node with only one child or no child
		if( (root->leftChild == NULL) || (root->rightChild == NULL) )
		{
			NodePointer temp = root->leftChild ? root->leftChild :
											root->rightChild;

			// No child case
			if (temp == NULL)
			{
				temp = root;
				root = NULL;
			}
			else // One child case
			*root = *temp; // Copy the contents of
							// the non-empty child
			free(temp);
		}
		else
		{
			// node with two children: Get the inorder
			// successor (smallest in the right subtree)
			NodePointer temp = minValueNode(root->rightChild);

			// Copy the inorder successor's data to this node
			root->data.key = temp->data.key;

			// Delete the inorder successor
			root->rightChild = deleteNode(root->rightChild, temp->data.key);
		}
	}

	// If the tree had only one node then return
	if (root == NULL)
	return root;

	// STEP 2: UPDATE HEIGHT OF THE CURRENT NODE
	root->height = 1 + getMaxValue(getTreeHeight(root->leftChild),
						getTreeHeight(root->rightChild));

	// STEP 3: GET THE BALANCE FACTOR OF THIS NODE (to
	// check whether this node became unbalanced)
	int balance = getBalance(root);

	// If this node becomes unbalanced, then there are 4 cases

	// Left Left Case
	if (balance > 1 && getBalance(root->leftChild) >= 0)
		return rightRotate(root);

	// Left Right Case
	if (balance > 1 && getBalance(root->leftChild) < 0)
	{
		root->leftChild = leftRotate(root->leftChild);
		return rightRotate(root);
	}

	// Right Right Case
	if (balance < -1 && getBalance(root->rightChild) <= 0)
		return leftRotate(root);

	// Right Left Case
	if (balance < -1 && getBalance(root->rightChild) > 0)
	{
		root->rightChild = rightRotate(root->rightChild);
		return leftRotate(root);
	}

	return root;
}
